﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Diagnostics;

namespace StateSpaceSearchCore
{
    struct StateInformation
    {
        public int GValue;
        public bool IsClosed;

        public StateInformation(int val = 0)
        {
            this.GValue = val;
            this.IsClosed = false;
        }

		public override string ToString()
		{
			return "cost=" + GValue;
		}
	}

    class SearchEngine
    {
        protected IHeap<int, State> openNodes;
        protected Dictionary<State, StateInformation> gValues;
        protected Dictionary<State, State> predecessor;
        public TimeSpan SearchTime;
        protected DateTime start;
        protected bool quiet = false;
        public List<State> Result;
        public int SolutionCost;
        protected List<StateCostPair> successors = new List<StateCostPair>();
		protected int branchingSum, branchingCount, maxGVal;

        protected virtual void addToClosedList(State state)
        {
            StateInformation f = gValues[state];
            f.IsClosed = true;
            gValues[state] = f;
			if (maxGVal < f.GValue)
				maxGVal = f.GValue;
        }

        protected virtual void addToOpenList(State s, int gValue, State pred)
        {
            if (!gValues.ContainsKey(s))
            {
                gValues.Add(s, new StateInformation(gValue));
                predecessor.Add(s, pred);
                openNodes.Insert(gValue, s);
                return;
            }
            if (gValues[s].GValue > gValue)
            {
                StateInformation f = gValues[s];
                f.GValue = gValue;
                f.IsClosed = false;
                gValues[s] = f;
                predecessor[s] = pred;
                openNodes.Insert(gValue, s);
                return;
            }
        }

        protected virtual void printSearchStats(bool quiet)
        {
            DateTime now = DateTime.Now;
            printMessage("Time elapsed: " + (now - start).TotalSeconds + " seconds", quiet);
			printMessage("Open nodes: " + openNodes.Size(), quiet);
			printMessage("Expanded nodes: " + gValues.Count, quiet);
			printMessage("Max GVal: " + maxGVal, quiet);
			printMessage("Speed: " + (gValues.Count / (now - start).TotalSeconds).ToString("0.00") + " states per second", quiet);
			printMessage("Average branching: " + (((double)branchingSum) / branchingCount).ToString("0.00"), quiet);
			printMessage("\n", quiet);
		}

        protected void printMessage(string message, bool quiet)
        {
            if (!quiet)
                Console.WriteLine(message);
        }

        public SearchEngine()
        {
            gValues = new Dictionary<State, StateInformation>();
            openNodes = new RedBlackTreeHeap<State>();
        }

        public virtual void Search(State s)
        {
			branchingCount = 1;
			branchingSum = 1;
			maxGVal = 0;

			printMessage("--- search started ---", quiet);
            State currentState = null;
            predecessor = new Dictionary<State, State>();
            gValues.Clear();
            openNodes.Clear();
            start = DateTime.Now;
            openNodes.Insert(0, s);
            gValues.Add(s, new StateInformation());
            predecessor.Add(s, null);
            int steps = -1;
            while (openNodes.Size() > 0)
            {
                steps++;

				if (steps % 10000 == 0)
                    printSearchStats(quiet);

                currentState = openNodes.RemoveMin();
                if (gValues[currentState].IsClosed)
                    continue;
                addToClosedList(currentState);
                if (currentState.IsFinal())
                {
                    DateTime end = DateTime.Now;
                    SearchTime = (end - start);
                    int GVAL = gValues[currentState].GValue;
                    printMessage("search ended in " + SearchTime.TotalSeconds + " seconds, plan length: " + GVAL, quiet);
                    printSearchStats(quiet);
                    this.Result = extractSolution(currentState);
                    this.SolutionCost = GVAL;
                    printMessage("--- search ended ---", quiet);
                    return;
                }
                int currentGValue = gValues[currentState].GValue;

                currentState.GetSuccessors(successors);
				branchingSum += successors.Count;
				branchingCount++;
                foreach (var item in successors)
                {
                    State state = item.State;
                    int gVal = currentGValue + item.Cost;
                    addToOpenList(state, gVal, currentState);
                }
            }
            printMessage("No solution exists.", quiet);
        }

        protected List<State> extractSolution(State state)
        {
            List<State> result = new List<State>();
            State current = state;
            while (current != null)
            {
                State pred = predecessor[current];
                if (pred == null)
                    break;
                result.Insert(0, current);
                current = pred;
            }
            return result;
        }

    }

}
